/**
 *
 * @version 2.0 (March 2019)
 *
 * Reference ISO-C11 Implementation of the LEDAcrypt PKC cipher using GCC built-ins.
 *
 * In alphabetical order:
 *
 * @author Marco Baldi <m.baldi@univpm.it>
 * @author Alessandro Barenghi <alessandro.barenghi@polimi.it>
 * @author Franco Chiaraluce <f.chiaraluce@univpm.it>
 * @author Gerardo Pelosi <gerardo.pelosi@polimi.it>
 * @author Paolo Santini <p.santini@pm.univpm.it>
 *
 * This code is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 **/

#include "mceliece_cca2_decrypt.h"
#include "H_Q_matrices_generation.h"
#include "gf2x_arith_mod_xPplusOne.h"
#include "bf_decoding.h"
#include "sha3.h"
#include "rng.h"
#include "constant_weight_codec.h"

#include <string.h> // memset(...), memcpy(....)
#include <assert.h>
#include <stdio.h>

/*----------------------------------------------------------------------------*/

static
int decrypt_McEliece (DIGIT decoded_err[],
                      DIGIT correct_codeword[],
                      AES_XOF_struct *mceliece_keys_expander,
                      int8_t rejections,
                      const unsigned char *const ctx)
{

   /* rebuild secret key values */

   POSITION_T HtrPosOnes[N0][DV];
   POSITION_T HPosOnes[N0][DV];
   POSITION_T QPosOnes[N0][M];

   POSITION_T LPosOnes[N0][DV*M];
   do {
   generateHPosOnes_HtrPosOnes(HPosOnes, HtrPosOnes,
                               mceliece_keys_expander);
   generateQsparse(QPosOnes, mceliece_keys_expander);
   for (int i = 0; i < N0; i++) {
        for (int j = 0; j< DV*M; j++) {
           LPosOnes[i][j]=INVALID_POS_VALUE;
        }
     }

     POSITION_T auxPosOnes[DV*M];
     unsigned char processedQOnes[N0] = {0};
     for (int colQ = 0; colQ < N0; colQ++) {
        for (int i = 0; i < N0; i++) {
           gf2x_mod_mul_sparse(DV*M, auxPosOnes,
                               DV, HPosOnes[i],
                               qBlockWeights[i][colQ], QPosOnes[i]+processedQOnes[i]);
           gf2x_mod_add_sparse(DV*M, LPosOnes[colQ],
                               DV*M, LPosOnes[colQ],
                               DV*M, auxPosOnes);
           processedQOnes[i] += qBlockWeights[i][colQ];
        }
     }
     rejections--;
   } while(rejections>=0);
   
   POSITION_T QtrPosOnes[N0][M];
   unsigned transposed_ones_idx[N0] = {0x00};
   for(unsigned source_row_idx=0; source_row_idx < N0 ; source_row_idx++) {
      int currQoneIdx = 0; // position in the column of QtrPosOnes[][...]
      int endQblockIdx = 0;
      for (int blockIdx = 0; blockIdx < N0; blockIdx++) {
         endQblockIdx += qBlockWeights[source_row_idx][blockIdx];
         for (; currQoneIdx < endQblockIdx; currQoneIdx++) {
            QtrPosOnes[blockIdx][transposed_ones_idx[blockIdx]] = (P -
                  QPosOnes[source_row_idx][currQoneIdx]) % P;
            transposed_ones_idx[blockIdx]++;
         }
      }
   }

   /* end rebuild secret key values */

   DIGIT codewordPoly[N0*NUM_DIGITS_GF2X_ELEMENT];
   memcpy(codewordPoly, ctx, N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B);
   for (unsigned i = 0; i < N0; i++) {
      gf2x_transpose_in_place(codewordPoly+i*NUM_DIGITS_GF2X_ELEMENT);
   }

   DIGIT privateSyndrome[NUM_DIGITS_GF2X_ELEMENT]; // privateSyndrome := yVar* Htr
   memset(privateSyndrome, 0x00, NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B);

   DIGIT aux[NUM_DIGITS_GF2X_ELEMENT];
   for(int i = 0; i < N0; i++) {
      gf2x_mod_mul_dense_to_sparse(aux,
                                   codewordPoly+i*NUM_DIGITS_GF2X_ELEMENT,
                                   LPosOnes[i],
                                   DV*M);
      gf2x_mod_add(privateSyndrome, privateSyndrome, aux);
   } // end for i
   gf2x_transpose_in_place(privateSyndrome);



   memset(decoded_err,0x00, N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B);
   /*perform syndrome decoding to obtain error vector */
   int ok;
   ok = bf_decoding(decoded_err,
                    (const POSITION_T (*)[DV])HtrPosOnes,
                    (const POSITION_T (*)[M])QtrPosOnes,
                    privateSyndrome
                   );

   if ( ok == 0 ) return 0;
   int err_weight = 0;
   for (int i =0 ;i < N0; i++){
       err_weight += population_count(decoded_err+(NUM_DIGITS_GF2X_ELEMENT*i));
   }
   if (err_weight != NUM_ERRORS_T) return 0;

   /* correct input codeword */
   for (unsigned i = 0; i < N0; i++) {
      gf2x_mod_add(correct_codeword+i*NUM_DIGITS_GF2X_ELEMENT,
                   (DIGIT *) ctx+i*NUM_DIGITS_GF2X_ELEMENT,
                   decoded_err+i*NUM_DIGITS_GF2X_ELEMENT);
   }
   return 1;
}
/*----------------------------------------------------------------------------*/
static
void char_left_bit_shift_n(const int length, uint8_t in[], const int amount)
{
   assert(amount < 8);
   if ( amount == 0 ) return;
   int j;
   uint8_t mask;
   mask = ~(((uint8_t)0x01 << (8 - amount)) - 1);
   for (j = 0 ; j < length-1 ; j++) {
      in[j] <<= amount;
      in[j] |=  (in[j+1] & mask) >> (8 - amount);
   }
   in[j] <<= amount;
} // end right_bit_shift_n



/*----------------------------------------------------------------------------*/
static
int poly_seq_into_bytestream(unsigned char output[],
                             const unsigned int byteOutputLength,
                             DIGIT zPoly[],
                             const unsigned int numPoly)
{
   DIGIT bitValue;
   unsigned int output_bit_cursor =  byteOutputLength*8-numPoly*P;

   if (NUM_BITS_GF2X_ELEMENT*numPoly > 8*byteOutputLength) return 0;

   for (int i =  0; i <numPoly; i++) {
      for (unsigned exponent = 0; exponent < NUM_BITS_GF2X_ELEMENT; exponent++) {
         bitValue = gf2x_get_coeff(zPoly+i*NUM_DIGITS_GF2X_ELEMENT, exponent);
         bitstream_write(output, 1, &output_bit_cursor, bitValue );
      } // end for exponent
   } // end for i
   int padsize = (K%8) ? 8-(K%8) : 0;
   char_left_bit_shift_n(byteOutputLength, output, padsize);
   return 1;
} // end poly_seq_into_bytestream

/*----------------------------------------------------------------------------*/

int decrypt_Kobara_Imai(unsigned char *const output,  //
                        unsigned long long * byteOutputLength,
                        AES_XOF_struct *mceliece_keys_expander,
                        int8_t rejections,
                        const unsigned long long clen,
                        const unsigned char *const
                        ctx // constituted by codeword || leftover
                       )
{
   DIGIT err[N0*NUM_DIGITS_GF2X_ELEMENT];
   DIGIT correctedCodeword[N0*NUM_DIGITS_GF2X_ELEMENT];

   /* first N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B bytes are the actual McE
    * ciphertext. Note: storage endiannes in BE hardware should flip bytes */
   memcpy(correctedCodeword,ctx,N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B);
   if (decrypt_McEliece (err, correctedCodeword,
                         mceliece_keys_expander,rejections, ctx) == 0 ) {
      fprintf(stderr,"Decode FAIL\n");
      return 0;
   }

   /* correctedCodeword now contains the correct codeword, iword is the first
    * portion, followed by syndrome turn back iword into a bytesequence */
   uint64_t paddedSequenceLen;
   if(clen <= N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B) {
     paddedSequenceLen =  (K+7)/8;
   } else {
     paddedSequenceLen = clen - N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B  -1 /*leftover dup!*/
                         + (K+7)/8;
   }
   unsigned char paddedOutput[paddedSequenceLen];
   memset(paddedOutput,0,paddedSequenceLen);
   poly_seq_into_bytestream(paddedOutput,(K+7)/8,
                            correctedCodeword,(N0-1));

   /* move back leftover padded string (if present) onto its position*/
   if (clen>N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B){
#if ((K%8)!=0)
   /* meld back byte split across iword and leftover. Recall that leftover is
    * built with leading zeroes, and output from iword has trailing zeroes
    * so no masking away is needed */
   paddedOutput[((K+7)/8)-1] |= ctx[N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B];
#endif

   int remainingToCopy= paddedSequenceLen - ((K+7)/8);
   memmove(paddedOutput+( (K+7)/8),
           ctx+N0*NUM_DIGITS_GF2X_ELEMENT*DIGIT_SIZE_B+1/*first leftover already managed*/,
           remainingToCopy);
   }


   unsigned char outputHash[HASH_BYTE_LENGTH] = {0};
   HASH_FUNCTION(paddedOutput,// param.s: input, inputLen, output
                 paddedSequenceLen,
                 outputHash);

   /* rebuild message hash ^ seed from error vector */
   unsigned char cwEncOutputBuffer[HASH_BYTE_LENGTH+CWENC_EXTRA_RANDOMNESS_LEN] = {0};
   constant_weight_to_binary_approximate(cwEncOutputBuffer, err);

   /* obtain back the PRNG seed */
   unsigned char secretSeed[TRNG_BYTE_LENGTH];
   for (int i = 0; i < TRNG_BYTE_LENGTH; i++) {
      secretSeed[i] = cwEncOutputBuffer[i] ^ outputHash[i];
   }

   /* test that the padding bytes of the seed are actually zero */
   for (int i = TRNG_BYTE_LENGTH; i < HASH_BYTE_LENGTH; i++) {
      if((cwEncOutputBuffer[i] ^ outputHash[i]) != 0){
         fprintf(stderr,"Nonzero TRNG pad \n");
         return 0;
      }
   }

   unsigned char prngSequence[paddedSequenceLen];
   memset (prngSequence,0x00,paddedSequenceLen);
   deterministic_random_byte_generator(prngSequence,sizeof(prngSequence),
                                       secretSeed,TRNG_BYTE_LENGTH);
   /* remove PRNG Pad from entire message */
   for (int i = 0;i < paddedSequenceLen;i++){
       paddedOutput[i] ^= prngSequence[i];
   }

   /*test if Kobara Imai constant, default to zero, matches */
   for (int i = 0; i < KOBARA_IMAI_CONSTANT_LENGTH_B; i++) {
      if(paddedOutput[i] != 0){
          fprintf(stderr,"KI const mismatch \n");
         return 0;
      }
   }
   /* retrieve message len, and set it */
   KI_LENGTH_FIELD_TYPE correctlySizedBytePtxLen ;

   memcpy(&(correctlySizedBytePtxLen),
          paddedOutput+KOBARA_IMAI_CONSTANT_LENGTH_B,
          KI_LENGTH_FIELD_SIZE);

   *byteOutputLength=correctlySizedBytePtxLen;
   /* copy message in output buffer */
   memcpy(output,
          paddedOutput+KOBARA_IMAI_CONSTANT_LENGTH_B+KI_LENGTH_FIELD_SIZE,
          correctlySizedBytePtxLen);
   return 1;
} // end decrypt_Kobara_Imai
